from ._core import (
    Agent as Agent,
)
from ._core import (
    Bool as Bool,
)
from ._core import (
    Byte as Byte,
)
from ._core import (
    Component as Component,
)
from ._core import (
    Coverage as Coverage,
)
from ._core import (
    Coverbin as Coverbin,
)
from ._core import (
    Covercross as Covercross,
)
from ._core import (
    Covergroup as Covergroup,
)
from ._core import (
    Coverpoint as Coverpoint,
)
from ._core import (
    Double as Double,
)
from ._core import (
    Driver as Driver,
)
from ._core import (
    Enum as Enum,
)
from ._core import (
    Env as Env,
)
from ._core import (
    Factory as Factory,
)
from ._core import (
    Fifo as Fifo,
)
from ._core import (
    Float as Float,
)
from ._core import (
    Fp16 as Fp16,
)
from ._core import (
    Fp32 as Fp32,
)
from ._core import (
    Fp64 as Fp64,
)
from ._core import (
    Half as Half,
)
from ._core import (
    IndexedScoreboard as IndexedScoreboard,
)
from ._core import (
    Int as Int,
)
from ._core import (
    Int8 as Int8,
)
from ._core import (
    Int16 as Int16,
)
from ._core import (
    Int32 as Int32,
)
from ._core import (
    Int64 as Int64,
)
from ._core import (
    List as List,
)
from ._core import (
    Log as Log,
)
from ._core import (
    Logic as Logic,
)
from ._core import (
    Memory as Memory,
)
from ._core import (
    Model as Model,
)
from ._core import (
    Monitor as Monitor,
)
from ._core import (
    Object as Object,
)
from ._core import (
    Phase as Phase,
)
from ._core import (
    PhaseManager as PhaseManager,
)
from ._core import (
    Port as Port,
)
from ._core import (
    Queue as Queue,
)
from ._core import (
    Scoreboard as Scoreboard,
)
from ._core import (
    Sequence as Sequence,
)
from ._core import (
    SequenceItem as SequenceItem,
)
from ._core import (
    Sequencer as Sequencer,
)
from ._core import (
    Struct as Struct,
)
from ._core import (
    Trace as Trace,
)
from ._core import (
    Transaction as Transaction,
)
from ._core import (
    Uint as Uint,
)
from ._core import (
    Uint8 as Uint8,
)
from ._core import (
    Uint16 as Uint16,
)
from ._core import (
    Uint32 as Uint32,
)
from ._core import (
    Uint64 as Uint64,
)
from ._core import (
    Visualization as Visualization,
)



# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Variable Class

import random
import warnings
from collections.abc import Callable
from typing import Any

from z3 import Int, Or

from .var import Var


class Enum(Var):

    def __copy__(self):
        """
        Copy the Logic - always make a copy to ensure randomness is preserved.

        Slightly unusual. Explicitly create an enum, then change type.
        This is to allow a user to subclass an enum with a pre-defined set of values.

        :return: Copied Var.
        :rtype: Var
        """
        new_obj = Enum(self.value,
                       self.values.copy(),
                       auto_random=self._auto_random_,
                       fmt=self._fmt_)
        new_obj._constraints_ = {
            k: v.copy() for k, v in self._constraints_.items()
        }
        new_obj.__class__ = self.__class__
        return  new_obj

    def __init__(
        self,
        *args,
        auto_random: bool = True,
        fmt : Callable [..., Any] = str
    ) -> None:
        """
        Initialize an enumeration variable.
        This class represents an enumeration variable that can take on a set of predefined values.
        The variable can be automatically randomized if `auto_random` is set to True.

        :param value: The initial value of the variable. It should be one of the values defined in `values`.
        :type value: Any
        :param values: A dictionary where keys are the names of the enumeration values and values are the corresponding values.
        :type values: dict[str, Any]
        :param auto_random: If True, the variable can be automatically randomized. Defaults to True.
        :type auto_random: bool
        :raises ValueError: If the provided `value` is not in the list of `values`.
        """
        if len(args) > 2 and self.__class__._deprecated_name_warning_:
            warnings.warn(
                "Passing 'name' as a positional argument is deprecated",
                DeprecationWarning,
                stacklevel=2
            )
            self.__class__._deprecated_name_warning_ = False

        value = args[-2]
        values = args[-1]

        # Define the values
        self.values = values
        for k, v in values.items():
            setattr(self, k, v)

        if value in values.keys():
            self.value = values[value]
        elif value in values.values():
            self.value = value
        else:
            raise ValueError(f"Value {value} is not in the list of values {values}")

        super().__init__(value, auto_random=auto_random, fmt=fmt)

        # Define a width - in case use in Struct
        self.width = max(values.values()).bit_length()

    def _cast_(self, other: Any) -> Any:
        """
        Cast the other value to the type of this variable's value.

        :param other: The value to cast.
        :type other: Any
        :return: The casted value.
        :rtype: Any
        """
        v = other.value if isinstance(other, type(self)) else other
        if v in self.values.keys():
            return self.values[v]
        elif v in self.values.values():
            return v
        else:
            raise ValueError(f"Value {v} is not in the list of values {self.values}")

    def _wrap_(self, result):
        """
        Wrap the result in an Var instance.

        :param result: The result to wrap.
        :type result: Any
        :return: An Var instance with the result.
        :rtype: Var
        """
        return type(self)(result, self.values, auto_random=self._auto_random_, fmt=self._fmt_)

    def _range_(self) -> tuple[Any, Any]:
        """
        Get the range of the variable.

        :return: A tuple containing the minimum and maximum values of the variable.
        :rtype: tuple[Any, Any]
        """
        return (min(self.values.values()), max(self.values.values()))

    def _z3_(self) -> Int:
        """
        Return the Z3 representation of the variable.
            - No need for constraint - better for performance
            - _random_value_ handles un-constrained situation
        :return: The Z3 representation of the variable.
        :rtype: BoolRef | IntNumRef | BitVecNumRef | RatNumRef
        """
        return Int(f"{self._idx_}")

    # Type Conversions
    def __str__(self) -> str:
        """
        Returns the string representation of the current instance.

        Iterates through the `values` dictionary and returns the key
        corresponding to the current `value`.

        :return: The string representation of the current instance.
        :rtype: str
        """
        for k, v in self.values.items():
            if v == self.value:
                return self._fmt_(k)

    def _random_value_(self, bounds: tuple[Any, Any] = None) -> Any:
        """
        Randomize the value of the variable.
        """
        if bounds is not None:
            values = []
            for v in self.values.values():
                if v >= min(bounds) and v <= max(bounds):
                    values.append(v)
            return random.choice(values)
        else:
            return random.choice(list(self.values.values()))

__all__ = ["Enum"]



# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Driver


from .component import Component
from .list import List


class Driver(Component):
    def __init__(self, name: str, parent: Component) -> None:
        """
        Initialize Driver.

        :param name: Name of the driver.
        :type name: str
        :param parent: Parent component.
        :type parent: Component
        """
        super().__init__(name, parent)
        self.seq_item_port = List()

__all__ = ["Driver"]



from ._vanilla import VanillaAgent, VanillaAgentCfg, VanillaDriver, VanillaEnv, VanillaEnvCfg, VanillaModel, VanillaMonitor, VanillaScoreboard, VanillaSequence, VanillaSequencer

__all__ = [
    "VanillaDriver",
    "VanillaMonitor",
    "VanillaSequencer",
    "VanillaSequence",
    "VanillaModel",
    "VanillaScoreboard",
    "VanillaAgentCfg",
    "VanillaAgent",
    "VanillaEnvCfg",
    "VanillaEnv",
]



# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Agent

from .component import Component


class Agent(Component):
    def __init__(self, name: str, parent: Component) -> None:
        """
        Initialize the avl_agent instance.

        :param name: Name of the agent instance
        :type name: str
        :param parent: Parent component
        :type parent: Component
        """
        super().__init__(name, parent)

__all__ = ["Agent"]



# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library List

from typing import Any

from cocotb.triggers import Event


class List(list):
    def __init__(self, *args: list[Any]) -> None:
        """
        Initializes the List.

        :param args: Initial elements of the list.
        :type args: list[Any]
        """
        super().__init__(args)
        self._push_event = Event()
        self._pop_event = Event()

    def append(self, data: Any) -> None:
        """
        Appends an element to the list and sets the event.

        :param data: The element to be appended.
        :type data: Any
        """
        super().append(data)
        self._push_event.set()

    def clear(self) -> None:
        """
        Clears the list and clears the event.
        """
        super().clear()
        self._pop_event.set()

    def extend(self, iterable: list[Any]) -> None:
        """
        Extends the list with elements from an iterable and sets the event.

        :param iterable: The iterable to extend the list with.
        :type iterable: list[Any]
        """
        super().extend(iterable)
        self._push_event.set()

    def insert(self, index: int, data: Any) -> None:
        """
        Inserts an element at a given position in the list and sets the event.

        :param index: Position at which to insert the element.
        :type index: int
        :param data: The element to be inserted.
        :type data: Any
        """
        super().insert(index, data)
        self._push_event.set()

    def pop(self, index: int = -1) -> Any:
        """
        Pops an element from the list at a given position.

        :param index: Position from which to pop the element. Defaults to -1 (last element).
        :type index: int
        :returns: The popped element.
        :rtype: Any
        """
        v = super().pop(index)
        self._pop_event.set()
        return v

    def remove(self, data: Any) -> None:
        """
        Removes the first occurrence of an element from the list and sets the event.

        :param data: The element to be removed.
        :type data: Any
        """
        super().remove(data)
        self._pop_event.set()

    async def blocking_pop(self) -> Any:
        """
        Pops an element from the list, blocking if the list is empty.

        :returns: The popped element.
        :rtype: Any
        """
        await self._push_event.wait()
        v = self.pop(0)
        if len(self) == 0:
            self._push_event.clear()
        return v

    async def blocking_get(self) -> Any:
        """
        Gets an element from the list, blocking if the list is empty.

        :returns: The retrieved element.
        :rtype: Any
        """
        return await self.blocking_pop()

Queue = List

__all__ = ["List"]



# Apheleia Verification Library Hierachical component
# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Coverpoint

from collections.abc import Callable
from typing import Any

import pandas as pd

from .component import Component
from .coverage import Coverage
from .covercross import Covercross
from .coverpoint import Coverpoint


class Covergroup(Component):
    def __init__(self, name: str, parent: Component = None) -> None:
        """
        Covergroup class for managing coverpoints and covercrosses.

        :param name: Name of the covergroup
        :type name: str
        :param parent: Parent component
        :type parent: Component, optional
        """
        super().__init__(name, parent)

        self.comment = None
        self.weight = 1
        self._cps_ = {}

        # Register
        Coverage().add_covergroup(self)

    def set_comment(self, comment: str) -> None:
        """
        Set a comment for the covergroup.

        :param comment: Comment for the covergroup
        :type comment: str
        """
        self.comment = comment

    def set_weight(self, weight: float) -> None:
        """
        Set the weight for the covergroup.

        :param weight: Weight of the covergroup
        :type weight: float
        """
        self.weight = weight

    def add_coverpoint(self, name: str, var: Callable[..., Any]) -> Coverpoint:
        """
        Add a coverpoint to the covergroup.

        :param name: Name of the coverpoint
        :type name: str
        :param var: Function that returns the value to be covered
        :type var: Callable[Any]
        :raises ValueError: If the coverpoint already exists
        :return: The created coverpoint
        :rtype: Coverpoint
        """
        if name not in self._cps_:
            self._cps_[name] = Coverpoint(name, self, var)
        else:
            raise ValueError(f"Coverpoint {name} already exists")
        return self._cps_[name]

    def add_covercross(self, name: str, *args: list[Coverpoint]) -> Covercross:
        """
        Add a covercross to the covergroup.

        :param name: Name of the covercross
        :type name: str
        :param args: Coverpoints to be crossed
        :type args: list[Coverpoint]
        :raises ValueError: If the covercross already exists
        :return: The created covercross
        :rtype: Covercross
        """
        if name not in self._cps_:
            self._cps_[name] = Covercross(name, self)
            self._cps_[name].add_points(*args)
        else:
            raise ValueError(f"Covercross {name} already exists")
        return self._cps_[name]

    def sample(self) -> None:
        """
        Sample all coverpoints in the covergroup
        """
        for cp in self._cps_.values():
            cp.sample()

    def clear(self) -> None:
        """
        Clear all coverpoints in the covergroup
        """
        for cp in self._cps_.values():
            cp.clear()

    def get_bins(self) -> tuple[int, int]:
        """
        Get the total and covered bins for the covergroup

        :return: A tuple (total, covered)
        :rtype: tuple[int, int]
        """
        total = 0
        covered = 0

        for cp in self._cps_.values():
            c, t = cp.get_bins()
            total += t
            covered += c

        return total, covered

    def get_coverage(self) -> float:
        """
        Get the coverage percentage for the covergroup

        :return: Coverage percentage
        :rtype: float
        """
        total, covered = self.get_bins()

        if total == 0:
            return float(0)
        else:
            return float(covered / total) * 100

    def report(self, full: bool = False) -> pd.DataFrame:
        """
        Generate a report for the covergroup

        :param full: If True, generate a detailed report
        :return: A pandas DataFrame with the report
        """
        retval = None
        for cp in self._cps_.values():
            if retval is None:
                retval = cp.report(full=full)
            else:
                retval = pd.concat([retval, cp.report(full=full)], ignore_index=True)

        retval.insert(0, "covergroup", self.name)
        retval = retval.fillna("")
        return retval


__all__ = ["Covergroup"]



# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Variable Class

from __future__ import annotations

import warnings
from collections.abc import Callable
from typing import Any

import numpy as np
from z3 import And, Real

from .var import Var


class Fp16(Var):
    def __init__(self, *args, auto_random: bool = True, fmt: Callable[..., float] = str) -> None:
        """
        Initialize an instance of the class.

        :param value: The value to be assigned to the instance.
        :type value: int
        :param auto_random: Flag to enable automatic randomization, defaults to True.
        :type auto_random: bool, optional
        :param fmt: The format to be used, defaults to hex.
        :type fmt: function, optional
        """
        super().__init__(*args, auto_random=auto_random, fmt=fmt)
        self._bits_ = np.uint16(0)

    def _cast_(self, other: Any) -> Any:
        """
        Cast the other value to the type of this variable's value.

        :param other: The value to cast.
        :type other: Any
        :return: The casted value.
        :rtype: Any
        """
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning, message="overflow encountered in cast")

            v = other.value if isinstance(other, type(self)) else other
            return np.float16(v)

    def _range_(self) -> tuple[int, int]:
        """
        Get the range of values that can be represented by this variable.

        :return: A tuple containing the minimum and maximum values.
        :rtype: tuple[int, int]
        """
        return (-np.finfo(self.value).max, np.finfo(self.value).max)

    def _z3_(self) -> Real:
        """
        Get the Z3 representation of the variable.
            - No need for constraint - better for performance
            - _random_value_ handles un-constrained situation
        :return: The Z3 FP representation of the variable.
        :rtype: FP
        """
        return Real(f"{self._idx_}")

    def _random_value_(self, bounds: tuple[float, float] = None) -> float:
        """
        Randomize the value of the variable.

        :param bounds: Optional bounds for the random value.
        :type bounds: tuple[float, float], optional
        :return: A random float value within the specified bounds or the maximum value.
        :rtype: float
        """
        if bounds is None:
            bounds = self._range_()
        x = np.random.uniform(min(bounds), max(bounds))
        return self._cast_(x)

    def to_bits(self) -> int:
        """
        Get the raw representation of the variable.

        :return: The raw value.
        :rtype: float
        """
        return int(self.value.view(type(self._bits_)))

    def from_bits(self, raw: int) -> None:
        """
        Convert the raw representation back to a float.

        :param raw: The raw value.
        :type raw: int
        """
        self.value = type(self._bits_)(int(raw)).view(type(self.value))

    # Bitwise
    def __and__(self, other): raise NotImplementedError("Bitwise operations are not supported for floating-point variables.")
    def __or__(self, other): raise NotImplementedError("Bitwise operations are not supported for floating-point variables.")
    def __xor__(self, other): raise NotImplementedError("Bitwise operations are not supported for floating-point variables.")
    def __lshift__(self, other): raise NotImplementedError("Bitwise operations are not supported for floating-point variables.")
    def __rshift__(self, other): raise NotImplementedError("Bitwise operations are not supported for floating-point variables.")
    def __iand__(self, other): raise NotImplementedError("Bitwise operations are not supported for floating-point variables.")
    def __ior__(self, other): raise NotImplementedError("Bitwise operations are not supported for floating-point variables.")
    def __ixor__(self, other): raise NotImplementedError("Bitwise operations are not supported for floating-point variables.")
    def __ilshift__(self, other): raise NotImplementedError("Bitwise operations are not supported for floating-point variables.")
    def __irshift__(self, other): raise NotImplementedError("Bitwise operations are not supported for floating-point variables.")
    def __rand__(self, other): raise NotImplementedError("Bitwise operations are not supported for floating-point variables.")
    def __ror__(self, other): raise NotImplementedError("Bitwise operations are not supported for floating-point variables.")
    def __rxor__(self, other): raise NotImplementedError("Bitwise operations are not supported for floating-point variables.")
    def __rlshift__(self, other): raise NotImplementedError("Bitwise operations are not supported for floating-point variables.")
    def __rrshift__(self, other): raise NotImplementedError("Bitwise operations are not supported for floating-point variables.")

    # Comparison - need to override to handle NaN and other cases
    def __eq__(self, other):
        other_val = self._cast_(other)
        return not (np.isnan(self.value) or np.isnan(other_val)) and self.value == other_val

    def __ne__(self, other):
        other_val = self._cast_(other)
        return np.isnan(self.value) or np.isnan(other_val) or self.value != other_val

    def __lt__(self, other):
        other_val = self._cast_(other)
        return not (np.isnan(self.value) or np.isnan(other_val)) and self.value < other_val

    def __le__(self, other):
        other_val = self._cast_(other)
        return not (np.isnan(self.value) or np.isnan(other_val)) and self.value <= other_val

    def __gt__(self, other):
        other_val = self._cast_(other)
        return not (np.isnan(self.value) or np.isnan(other_val)) and self.value > other_val

    def __ge__(self, other):
        other_val = self._cast_(other)
        return not (np.isnan(self.value) or np.isnan(other_val)) and self.value >= other_val

class Fp32(Fp16):
    def __init__(self, *args, auto_random: bool = True, fmt: Callable[..., float] = str) -> None:
        """
        Initialize an instance of the class.

        :param value: The value to be assigned to the instance.
        :type value: int
        :param auto_random: Flag to enable automatic randomization, defaults to True.
        :type auto_random: bool, optional
        :param fmt: The format to be used, defaults to hex.
        :type fmt: function, optional
        """
        super().__init__(*args, auto_random=auto_random)
        self._bits_ = np.uint32(0)

    def _cast_(self, other: Any) -> Any:
        """
        Cast the other value to the type of this variable's value.

        :param other: The value to cast.
        :type other: Any
        :return: The casted value.
        :rtype: Any
        """
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning, message="overflow encountered in cast")

            v = other.value if isinstance(other, type(self)) else other
            return np.float32(v)

class Fp64(Fp16):
    def __init__(self, *args, auto_random: bool = True, fmt: Callable[..., float] = str) -> None:
        """
        Initialize an instance of the class.

        :param value: The value to be assigned to the instance.
        :type value: int
        :param auto_random: Flag to enable automatic randomization, defaults to True.
        :type auto_random: bool, optional
        :param fmt: The format to be used, defaults to hex.
        :type fmt: function, optional
        """
        super().__init__(*args, auto_random=auto_random)
        self._bits_ = np.uint64(0)

    def _cast_(self, other: Any) -> Any:
        """
        Cast the other value to the type of this variable's value.

        :param other: The value to cast.
        :type other: Any
        :return: The casted value.
        :rtype: Any
        """
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning, message="overflow encountered in cast")

            v = other.value if isinstance(other, type(self)) else other
            return np.float64(v)

    def _range_(self) -> tuple[int, int]:
        """
        Get the range of values that can be represented by this variable.

        :return: A tuple containing the minimum and maximum values.
        :rtype: tuple[int, int]
        """
        return (-1e100, 1e100) # Reduced to allow randomization

Half = Fp16
Float = Fp32
Double = Fp64

__all__ = ["Fp16", "Fp32", "Fp64", "Half", "Float", "Double"]



# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Variable Class

from __future__ import annotations

from collections.abc import Callable
from typing import Any

from z3 import BitVec

from .var import Var


class Logic(Var):

    def __copy__(self):
        """
        Copy the Logic - always make a copy to ensure randomness is preserved.

        :return: Copied Var.
        :rtype: Var
        """
        new_obj = Logic(self.value, auto_random=self._auto_random_, fmt=self._fmt_, width=self.width)
        new_obj._constraints_ = {
            k: v.copy() for k, v in self._constraints_.items()
        }
        return  new_obj

    def __init__(
        self,
        *args,
        auto_random: bool = True,
        fmt: Callable[..., int] = hex,
        width: int = 32
    ) -> None:
        """
        Initialize an instance of the class.

        :param value: The initial value of the variable.
        :type value: any
        :param auto_random: Indicates if the variable should be automatically randomized, defaults to True.
        :type auto_random: bool, optional
        :param fmt: The format of the variable, defaults to hex.
        :type fmt: type, optional
        :param width: The width of the variable in bits, defaults to 32.
        :type width: int, optional
        :raises ValueError: If the width is not a positive integer.
        """
        if not isinstance(width, int) or width <= 0:
            raise ValueError("Width must be a positive integer.")
        self.width = int(width)

        super().__init__(*args, auto_random=auto_random, fmt=fmt)

    def _cast_(self, other: Any) -> int:
        """
        Cast the value to the appropriate type based on the width of the variable.

        :param other: The value to be cast.
        :type other: Any
        :return: The casted value.
        :rtype: int
        """
        v = other.value if isinstance(other, Logic) else other
        return int(v & self._range_()[1])

    def _wrap_(self, result : Any) -> Logic:
        """
        Wrap the result in an Logic instance.

        :param result: The result to be wrapped.
        :type result: Any
        :return: An instance of Logic with the result.
        :rtype: Logic
        """
        return type(self)(result, auto_random=self._auto_random_, fmt=self._fmt_, width=self.width)

    def _range_(self) -> tuple[int, int]:
        """
        Get the range of values that can be represented by this variable.

        :return: A tuple containing the minimum and maximum values.
        :rtype: tuple[int, int]
        """
        return (0, (1 << self.width) - 1)

    def _z3_(self) -> BitVec:
        """
        Get the Z3 representation of the variable.

        :return: The Z3 BitVec representation of the variable.
        :rtype: z3.BitVecRef
        """
        return BitVec(f"{self._idx_}", self.width)

__all__ = ["Logic"]



# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Scoreboard

from .component import Component
from .list import List


class Scoreboard(Component):
    def __init__(self, name: str, parent: Component) -> None:
        """
        Initializes the scoreboard component.

        :param name: Name of the scoreboard.
        :type name: str
        :param parent: Parent component.
        :type parent: Component
        """

        super().__init__(name, parent)
        self.before_port = List()
        self.before_item = None
        self.after_port = List()
        self.after_item = None
        self.compare_count = 0
        self.min_compare_count = 0
        self.verbose = False

    def set_verbose(self, verbose: bool) -> None:
        """
        Sets the verbosity of the scoreboard.

        :param verbose: Verbosity flag.
        """
        self.verbose = verbose

    def set_min_compare_count(self, count: int) -> None:
        """
        Sets the minimum number of comparisons required.

        :param count: Minimum number of comparisons.
        """
        self.min_compare_count = count

    async def run_phase(self) -> None:
        """
        Runs the scoreboard phase.
        """
        while True:
            self.before_item = await self.before_port.blocking_pop()
            self.after_item = await self.after_port.blocking_pop()

            self.before_item.compare(self.after_item, verbose=self.verbose, bidirectional=True)

            self.compare_count += 1

            self.after_item = None
            self.before_item = None

    async def report_phase(self) -> None:
        """
        Reports the results of the scoreboard.
        """
        if len(self.before_port) + int(self.before_item is not None) > 0 or len(
            self.after_port
        ) + int(self.after_item is not None):
            self.error(
                f"Outstanding items: before_port={len(self.before_port) + int(self.before_item is not None)} after_port={len(self.after_port) + int(self.after_item is not None)}"
            )

        if self.compare_count < self.min_compare_count:
            self.error(
                f"Not enough items compared: {self.compare_count} < {self.min_compare_count}"
            )

        if self.verbose:
            self.info(f"Compared {self.compare_count} items")


__all__ = ["Scoreboard"]



# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Hierachical component

from __future__ import annotations

from typing import Any

import cocotb
from cocotb.triggers import NullTrigger

from .object import Object
from .phase import Phase
from .phase_manager import PhaseManager
from .visualization import Visualization


class Component(Object):

    def __deepcopy__(self, memo):
        """
        Deep copy the Component - to avoid recursion just return the instance.

        :param memo: Dictionary to keep track of already copied objects.
        :type memo: dict
        :return: Deep copied Component.
        :rtype: Component
        """
        return self

    def __init__(self, name: str, parent: Component) -> None:
        """
        Initialize Component.

        :param name: Name of the component.
        :type name: str
        :param parent: Parent component.
        :type parent: Component
        """
        super().__init__(name, parent)

        # Parent / Child
        self._children_ = []
        if parent is not None:
            parent.add_child(self)

        Visualization.add_component(self)

        # Phase management
        # Sync to ensure all children complete their hierarchical function calls
        # before the parent does.
        self._hierarchical_sync_ = cocotb.triggers.Event()

    async def _hierarchical_func_(self, fn_name: str, *args: Any, **kwargs: Any) -> Any:
        """
        Execute a hierarchical function on all child components.

        :param fn_name: Name of the function to execute.
        :type fn_name: str
        :param args: Variable length argument list.
        :param kwargs: Arbitrary keyword arguments.
        """
        phase = PhaseManager._current
        if phase.top_down:
            fn = getattr(self, fn_name, None)
            if fn is not None:
                await fn(*args, **kwargs)

        for c in self.get_children():
            cocotb.start_soon(c._hierarchical_func_(fn_name, *args, **kwargs))
            await NullTrigger()
            if not phase.top_down:
                await c._hierarchical_sync_.wait()
                c._hierarchical_sync_.clear()

        if not phase.top_down:
            self._hierarchical_sync_.set()
            fn = getattr(self, fn_name, None)
            if fn is not None:
                await fn(*args, **kwargs)

    def add_child(self, child: Component) -> None:
        """
        Add a child component.

        :param child: Child component to add.
        :type child: Component
        """
        if child not in self._children_:
            self._children_.append(child)

    def get_child(self, name: str) -> Component:
        """
        Get a child component by name.

        :param name: Name of the child component.
        :type name: str
        :return: Child component.
        :rtype: Component
        """
        for c in self._children_:
            if c.name == name:
                return c
        return None

    def get_num_children(self) -> int:
        """
        Get the number of child components.

        :return: Number of child components.
        :rtype: int
        """
        return len(self._children_)

    def get_children(self) -> list[Component]:
        """
        Get all child components.

        :return: List of child components.
        :rtype: list
        """
        return self._children_

    async def start(self) -> None:
        """
        Start the component and execute its phases.
        """
        while PhaseManager._current is not None:
            fn_name = f"{PhaseManager._current.name.lower()}_phase"
            await self._hierarchical_func_(fn_name)
            await PhaseManager._current.wait_for_objections()
            PhaseManager.next()

    def raise_objection(self, phase: Phase = None, obj: Object = None) -> None:
        """
        Raise an objection for the current phase.

        :param phase: Phase to raise objection for.
        :type phase: Phase
        :param obj: Object raising the objection.
        :type obj: Component
        """
        if phase is None:
            phase = PhaseManager._current

        if obj is None:
            obj = self

        phase.raise_objection(obj)

    def drop_objection(self, phase: Phase = None, obj: Object = None) -> None:
        """
        Drop an objection for the current phase.

        :param phase: Phase to drop objection for.
        :type phase: Phase
        :param obj: Object dropping the objection.
        :type obj: Component
        """
        if phase is None:
            phase = PhaseManager._current

        if obj is None:
            obj = self

        phase.drop_objection(obj)


__all__ = ["Component"]



# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Monitor


from .component import Component
from .port import Port


class Monitor(Component):
    def __init__(self, name: str, parent: Component) -> None:
        """
        Initialize the Monitor instance.

        :param name: Name of the monitor.
        :type name: str
        :param parent: Parent component.
        :type parent: Component
        """
        super().__init__(name, parent)
        self.item_export = Port("item_export", self)

__all__ = ["Monitor"]



# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Memory Model

import bincopy
import pandas as pd


class Memory:

    def __init__(self, width : int = 32) -> None:
        """
        Initialize the memory model.

        :param width: Width of the memory in bits (default is 32).
        :type width: int
        :raises ValueError: If width is not a positive integer.
        """
        if width <= 0 or width % 8 != 0:
            raise ValueError("Width must be a positive integer and a multiple of 8.")

        self.width = width
        self.ranges = []
        self.memory = {}
        self.endianness = 'little'
        self.init_fn = lambda address : 0

    def _align_address_(self, address: int) -> int:
        """
        Align the address to the memory width.

        :param address: Address to align.
        :type address: int
        :return: Aligned address.
        :rtype: int
        """
        return address & ~(self.width // 8 - 1)

    def _check_address_(self, address: int) -> None:
        """
        Check if the address is valid.

        :param address: Address to check.
        :type address: int
        :raises ValueError: If address is not a positive integer.
        """
        for r in self.ranges:
            if address >= r[0] and address < r[1]:
                return True

        self.miss(address)
        return False

    def _get_byte_(self, address: int) -> int:
        """
        Get the byte at the specified address.

        :param address: Address to get the byte from.
        :type address: int
        :return: Byte value at the specified address.
        :rtype: int
        :raises KeyError: If address is not in memory.
        """
        if address not in self.memory:
            self.memory[address] = self.init_fn(address)
        return self.memory[address]

    def set_init_fn(self, fn : callable) -> None:
        """
        Set the initialization policy for the memory

        Defined as a lambda function that takes an address and returns a value.

        :param fn: Function to initialize memory at a given address.
        :type fn: callable
        :raises ValueError: If fn is not callable.
        :example:
            memory.set_init_fn(lambda address: 0)
        """
        self.init_fn = fn

    def set_endianness(self, endianness: str) -> None:
        """
        Set the endianness of the memory (little / big).

        :param endianness: Endianness, can be 'little' or 'big'.
        :type endianness: str
        :raises ValueError: If endianness is not 'little' or 'big'.
        """
        if endianness not in ['little', 'big']:
            raise ValueError("Endianness must be either 'little' or 'big'.")
        self.endianness = endianness

    def add_range(self, start: int, end: int) -> None:
        """
        Add a memory range.

        :param start: Start address of the memory range.
        :type start: int
        :param end: End address of the memory range.
        :type end: int
        :raises ValueError: If start address is not less than end address.
        """
        if start >= end:
            raise ValueError("Start address must be less than end address.")
        self.ranges.append((start, end))

    def miss(self, address : int) -> None:
        """
        Action to be taken when a memory miss occurs.
        This method raises a ValueError indicating that the address was not found in memory.
        Expect user to override this method to implement custom behavior.

        :param address: Address that caused the miss.
        :type address: int
        :raises ValueError: Always raised to indicate a memory miss.
        """
        raise KeyError(f"Miss at {address}")

    def read(self, address: int, num_bytes : int = None) -> int:
        """
        Read a value from the memory at the specified address.

        Calls miss() if the address is not found in memory.

        :param address: Address to read from.
        :type address: int
        :return: Value at the specified address.
        :rtype: int
        """
        if num_bytes is None:
            num_bytes = self.width // 8

        self._check_address_(address)
        self._check_address_(address + num_bytes - 1)

        data = bytearray()
        for i in range(num_bytes):
            data.append(self._get_byte_(address + i))

        return int.from_bytes(data, self.endianness)

    def write(self, address: int, value: int, num_bytes : int = None, strobe : int = None) -> None:
        """
        Write a value to the memory at the specified address.

        Calls miss() if the address is not found in memory.

        :param address: Address to write to.
        :type address: int
        :param value: Value to write.
        :type value: int
        :param num_bytes: Number of bytes to write (default is width // 8).
        :type num_bytes: int, optional
        :param strobe: Strobe signal
        :type strobe: int, optional
        """
        if num_bytes is None:
            num_bytes = self.width // 8

        if strobe is None:
            strobe = (1 << num_bytes) - 1

        self._check_address_(address)
        self._check_address_(address + num_bytes - 1)

        data = value.to_bytes(num_bytes, self.endianness)
        for i in range(num_bytes):
            offset = i if self.endianness == 'little' else num_bytes - 1 - i
            if strobe & (1 << offset):
                self.memory[address + i] = data[i]

    def export_to_file(self, filename: str, fmt : str = None) -> None:
        """
        Export memory contents to a file.

        If fmt is not specified, it will be inferred from the file extension.

        :param filename: Path to the file where memory contents will be saved.
        :type filename: str
        :param fmt: Format of the file, can be 'vhex' or 'vbin'.
        :type fmt: str, optional
        :raises ValueError: If format is not supported.
        """

        def exists(address: int):
            """
            Pad the whole memory width
            """
            exists = False
            for i in range(address, address + self.width // 8):
                if i in self.memory:
                    exists = True
                    break

            self.read(self._align_address_(address))
            return exists

        def verilog(filename: str, fmt : str) -> None:
            """
            Export memory contents to a verilog hex file.
            """

            with open(filename, 'w') as f:
                for r in self.ranges:
                    new_address = True
                    for a in range(r[0], r[1], self.width // 8):
                        if exists(a):
                            if new_address:
                                f.write(f"@{a:04x}\n")
                                new_address = False

                            if fmt == "vhex":
                                f.write(f"{self.read(a, self.width // 8):0{self.width // 4}x}\n")
                            elif fmt == "vbin":
                                f.write(f"{self.read(a, self.width // 8):0{self.width}b}\n")
                        else:
                            new_address = True

        def pandas(filename: str, fmt : str) -> None:
            """
            Export memory contents to a pandas DataFrame and save to file.
            """
            df = pd.DataFrame(list(self.memory.items()), columns=["addr", "data"])

            if fmt == "csv":
                df.to_csv(filename, index=False)
            elif fmt == "json":
                df.to_json(filename, orient='records', lines=True)
            else:
                raise ValueError(f"Unsupported file format: {fmt}")

        def bcopy(filename: str, fmt : str) -> None:
            """
            Export memory contents using bincopy.
            """
            bf = bincopy.BinFile()
            for a,d in self.memory.items():
                bf[a] = d

            if fmt in ["ihex", "hex", "ihx"]:
                with open(filename, 'w') as f:
                    f.write(bf.as_ihex())
            elif fmt == "srec":
                with open(filename, 'w') as f:
                    f.write(bf.as_srec())
            elif fmt == "ti-txt":
                with open(filename, 'w') as f:
                    f.write(bf.as_ti_txt())
            elif fmt == "vmem":
                with open(filename, 'w') as f:
                    f.write(bf.as_verilog_vmem())
            else:
                raise ValueError(f"Unsupported file format: {fmt}")

        # Default format from file extension
        if fmt is None:
            fmt = filename.split('.')[-1]

        if fmt in ["vhex", "vbin"]:
            verilog(filename, fmt=fmt)
        elif fmt in ["csv", "json"]:
            pandas(filename, fmt=fmt)
        else:
            try:
                bcopy(filename, fmt=fmt)
            except Exception as e:
                raise ValueError(f"Unsupported file format: {fmt}") from e

    def import_from_file(self, filename: str, fmt : str = None) -> None:
        """
        Load memory contents from a file.

        If fmt is not specified, it will be inferred from the file extension.

        :param filename: Path to the file containing memory contents.
        :type filename: str
        :raises FileNotFoundError: If the file does not exist.
        """
        def verilog(filename: str, fmt : str) -> None:
            """
            Load memory contents from a verilog hex file.
            """
            addr = self.ranges[0][0] if self.ranges else 0

            with open(filename) as f:
                for raw in f:
                    line = raw.strip()
                    if not line:
                        continue

                    # Comments (//, ;, #)
                    for sep in ('//', ';', '#'):
                        if sep in line:
                            line = line.split(sep, 1)[0]
                    line = line.strip()
                    if not line:
                        continue

                    # New base address
                    if line.startswith('@'):
                        try:
                            addr = int(line[1:], 16)
                        except Exception as e:
                            raise ValueError(f"Invalid address marker: {line!r}") from e
                        continue

                    # Tokens separated by whitespace
                    t = line.replace(" ", "")
                    if t.startswith('0x') or t.startswith('0b'):
                        t = t[2:]

                    if fmt == "vhex":
                        bytes = int(t, 16).to_bytes(len(t) // 2, self.endianness)
                        for i in range(len(bytes)):
                            self.write(addr+i, bytes[i], 1)
                    elif fmt == "vbin":
                        bytes = int(t, 2).to_bytes(len(t) // 8, self.endianness)
                        for i in range(len(bytes)):
                            self.write(addr+i, bytes[i], 1)
                    else:
                        raise ValueError(f"Unsupported file format: {fmt}")
                    addr += self.width // 8

        def pandas(filename: str, fmt : str) -> None:
            """
            Export memory contents to a pandas DataFrame and save to file.
            """
            if fmt == "csv":
                df = pd.read_csv(filename)
            elif fmt == "json":
                df = pd.read_json(filename, orient='records', lines=True)
            else:
                raise ValueError(f"Unsupported file format: {fmt}")

            self.memory = dict(zip(df["addr"], df["data"], strict=False))

        def bcopy(filename: str, fmt : str) -> None:
            """
            Import memory contents using bincopy.
            """
            bf = bincopy.BinFile(filename)

            for start_address, data in bf.segments:
                for offset, byte_val in enumerate(data):
                    addr = start_address + offset
                    self.memory[addr] = byte_val

        # Default format from file extension
        if fmt is None:
            fmt = filename.split('.')[-1]

        if fmt in ["vhex", "vbin"]:
            verilog(filename, fmt=fmt)
        elif fmt in ["csv", "json"]:
            pandas(filename, fmt=fmt)
        else:
            try:
                bcopy(filename, fmt=fmt)
            except Exception as e:
                raise ValueError(f"Unsupported file format: {fmt}") from e

__all__ = ["Memory"]



# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Variable Class

from __future__ import annotations

from collections.abc import Callable
from typing import Any

from .uint import Uint
from .var import Var


class Int(Uint):
    def __copy__(self):
        """
        Copy the Logic - always make a copy to ensure randomness is preserved.

        :return: Copied Var.
        :rtype: Var
        """
        new_obj = Int(self.value, auto_random=self._auto_random_, fmt=self._fmt_, width=self.width)
        new_obj._constraints_ = {
            k: v.copy() for k, v in self._constraints_.items()
        }
        new_obj.__class__ = self.__class__
        return  new_obj

    def __init__(
        self,
        *args,
        auto_random: bool = True,
        fmt: Callable[..., int] = str,
        width: int = 32
    ) -> None:
        """
        Initialize an instance of the class.

        :param value: The value to be assigned to the instance.
        :type value: int
        :param auto_random: Flag to enable automatic randomization, defaults to True.
        :type auto_random: bool, optional
        :param fmt: The format to be used, defaults to str.
        :type fmt: function, optional
        :param width: The width of the variable in bits, defaults to 32.
        :type width: int, optional
        :raises ValueError: If the width is not a positive integer.
        """
        super().__init__(*args, auto_random=auto_random, fmt=fmt, width=width)

    def _cast_(self, other: Any) -> int:
        """
        Cast the value to the appropriate type based on the width of the variable.

        :param other: The value to be cast.
        :type other: Any
        :return: The casted value.
        :rtype: int
        """
        v = other.value if isinstance(other, Var) else other
        (_min_, _max_) = self._range_()

        return int((v - _min_) % (1 << self.width)) + _min_

    def _range_(self) -> tuple[int, int]:
        """
        Get the range of values that can be represented by this variable.

        :return: A tuple containing the minimum and maximum values.
        :rtype: tuple[int, int]
        """
        return (-(1 << (self.width - 1)), (1 << (self.width - 1)) - 1)

class Int8(Int):
    def __init__(
        self, *args, auto_random: bool = True, fmt: Callable[..., int] = str
    ) -> None:
        """
        Initialize an instance of the class.

        :param value: The value to be assigned to the instance.
        :type value: int
        :param auto_random: Flag to enable automatic randomization, defaults to True.
        :type auto_random: bool, optional
        :param fmt: The format to be used, defaults to str.
        :type fmt: function, optional
        """
        super().__init__(*args, auto_random=auto_random, fmt=fmt, width=8)

    def _wrap_(self, result : Any) -> Int8:
        """
        Wrap the result in an avl_logic instance.

        :param result: The result to be wrapped.
        :type result: Any
        :return: An instance of avl_logic with the result.
        :rtype: avl_logic
        """
        return type(self)(result, auto_random=self._auto_random_, fmt=self._fmt_)

class Int16(Int):
    def __init__(
        self, *args, auto_random: bool = True, fmt: Callable[..., int] = str
    ) -> None:
        """
        Initialize an instance of the class.

        :param value: The value to be assigned to the instance.
        :type value: int
        :param auto_random: Flag to enable automatic randomization, defaults to True.
        :type auto_random: bool, optional
        :param fmt: The format to be used, defaults to str.
        :type fmt: function, optional
        """
        super().__init__(*args, auto_random=auto_random, fmt=fmt, width=16)

    def _wrap_(self, result : Any) -> Int16:
        """
        Wrap the result in an avl_logic instance.

        :param result: The result to be wrapped.
        :type result: Any
        :return: An instance of avl_logic with the result.
        :rtype: avl_logic
        """
        return type(self)(result, auto_random=self._auto_random_, fmt=self._fmt_)

class Int32(Int):
    def __init__(
        self, *args, auto_random: bool = True, fmt: Callable[..., int] = str
    ) -> None:
        """
        Initialize an instance of the class.

        :param value: The value to be assigned to the instance.
        :type value: int
        :param auto_random: Flag to enable automatic randomization, defaults to True.
        :type auto_random: bool, optional
        :param fmt: The format to be used, defaults to str.
        :type fmt: function, optional
        """
        super().__init__(*args, auto_random=auto_random, fmt=fmt, width=32)

    def _wrap_(self, result : Any) -> Int32:
        """
        Wrap the result in an avl_logic instance.

        :param result: The result to be wrapped.
        :type result: Any
        :return: An instance of avl_logic with the result.
        :rtype: avl_logic
        """
        return type(self)(result, auto_random=self._auto_random_, fmt=self._fmt_)

class Int64(Int):
    def __init__(
        self, *args, auto_random: bool = True, fmt: Callable[..., int] = str
    ) -> None:
        """
        Initialize an instance of the class.

        :param value: The value to be assigned to the instance.
        :type value: int
        :param auto_random: Flag to enable automatic randomization, defaults to True.
        :type auto_random: bool, optional
        :param fmt: The format to be used, defaults to str.
        :type fmt: function, optional
        """
        super().__init__(*args, auto_random=auto_random, fmt=fmt, width=64)

    def _wrap_(self, result : Any) -> Int64:
        """
        Wrap the result in an avl_logic instance.

        :param result: The result to be wrapped.
        :type result: Any
        :return: An instance of avl_logic with the result.
        :rtype: avl_logic
        """
        return type(self)(result, auto_random=self._auto_random_, fmt=self._fmt_)

Byte = Int8

__all__ = ["Int", "Int8", "Int16", "Int32", "Int64", "Byte"]



# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Transacation

from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from cocotb.triggers import Event
from cocotb.utils import get_sim_time

from .object import Object

if TYPE_CHECKING:
    from .component import Component

class Transaction(Object):
    def __init__(self, name: str, parent: Component) -> None:
        """
        Initialize a new Transaction.

        :param name: The name of the transaction.
        :type name: str
        """
        super().__init__(name, parent)
        self._id_ = -1
        self._events_ = {}

    def set_id(self, id: int) -> None:
        """
        Set the ID of the transaction.

        :param id: The ID to set.
        :type id: int
        """
        self._id_ = id

    def get_id(self) -> int:
        """
        Get the ID of the transaction.

        :return: The ID of the transaction.
        :rtype: int
        """
        return self._id_

    def add_event(self, name: str, callback: Callable[..., Any] = None) -> None:
        """
        Add an event to the transaction.

        :param name: The name of the event.
        :type name: str
        :param callback: The callback function to be called when the event is set.
        :type callback: function or None
        """
        if name not in self._events_:
            self._events_[name] = [0, Event(), []]

        if callback is not None:
            self._events_[name][2].append(callback)

    def get_event(self, name: str) -> None:
        """
        Get an event by name.

        :param name: The name of the event.
        :type name: str
        :return: The event details or None if the event does not exist.
        :rtype: list or None
        """
        if name in self._events_:
            return self._events_[name]
        else:
            return None

    def set_event(self, name: str, *args: list[Any], **kwargs: list[Any]) -> None:
        """
        Set an event and trigger its callbacks.

        :param name: The name of the event.
        :type name: str
        :param args: Additional arguments for the callback.
        :param kwargs: Additional keyword arguments for the callback.
        """
        if "unit" in kwargs:
            self._events_[name][0] = get_sim_time(unit=kwargs["unit"])
        else:
            self._events_[name][0] = get_sim_time(unit="ns")

        self._events_[name][1].set()

        for cb in self._events_[name][2]:
            if cb is not None:
                cb(*args, **kwargs)

    async def wait_on_event(self, name: str) -> None:
        """
        Wait for an event to be set.

        :param name: The name of the event.
        :type name: str
        """
        await self._events_[name][1].wait()


__all__ = ["Transaction"]



# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Coverage Container

from __future__ import annotations

import atexit
from typing import TYPE_CHECKING

import pandas as pd

if TYPE_CHECKING:
    from .covergroup import CoverGroup

class Coverage:
    _instance = None

    def __new__(cls) -> Coverage:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        """
        Initialize the coverage container
        This method initializes the coverage container and registers the exit handler
        to export coverage data when the program exits.
        It checks if the `_cg_` attribute exists; if not, it initializes it as an empty list.
        If the `_cg_` attribute already exists, it does nothing to avoid re-initialization.
        """
        if hasattr(self, "_cg_"):
            return

        # Cover Groups
        self._cg_ = []

        # Register function to export coverage at exit
        atexit.register(self.__at_exit__)

    def __at_exit__(self) -> None:
        """
        Export coverage data at exit
        """
        if len(self._cg_) == 0:
            return

        df = None
        for cg in self._cg_:
            if df is None:
                df = cg.report(full=True)
            else:
                df = pd.concat([df, cg.report(full=True)], ignore_index=True)

        df.to_json("coverage.json", mode="w", orient="records")

    def add_covergroup(self, cg: CoverGroup) -> None:
        """
        Add a covergroup to the coverage container

        :param cg: The covergroup to add
        """
        self._cg_.append(cg)

    def remove_covergroup(self, cg: CoverGroup) -> None:
        """
        Remove a covergroup from the coverage container

        :param cg: The covergroup to remove
        """
        self._cg_.remove(cg)


__all__ = ["Coverage"]



# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Trace

import os

import cocotb
import pandas as pd

from .component import Component
from .factory import Factory
from .list import List


class Trace(Component):
    def __init__(self, name: str, parent: Component) -> None:
        """
        Initialize the AVL model component.

        :param name: Name of the model.
        :type name: str
        :param parent: Parent component of the model.
        :type parent: Component
        """
        super().__init__(name, parent)
        self.item_port = List()

        # Interval to flush trace data to file
        self.flush_interval = Factory.get_variable(f"{self.get_full_name()}.flush_interval", 100)

        # Trace name
        self.tracefile = Factory.get_variable(f"{self.get_full_name()}.tracefile", f"{self.get_full_name()}.csv")

        # Delete old trace file if it exists
        if os.path.exists(self.tracefile):
            os.remove(self.tracefile)

        # User defined columns for trace data
        # If None, all public attributes of item will be used
        self.columns = None

        # Pandas DataFrame to hold trace data
        self.df = self.create_dataframe()

    def create_dataframe(self) -> pd.DataFrame:
        """
        Create a new DataFrame with the specified columns.

        :return: A new DataFrame with the specified columns.
        :rtype: pd.DataFrame
        """
        if self.columns is not None:
            if '_path_' not in self.columns:
                self.columns.insert(0, '_path_')

            if '_timestamp_' not in self.columns:
                self.columns.insert(0, '_timestamp_')

            return pd.DataFrame(columns=self.columns)
        else:
            return None

    def flush(self) -> None:
        """
        Flush the trace data to a file or output.

        This method can be overridden to implement custom flush behavior.
        """
        if self.df is None:
            return

        self.debug(f"Flushing trace data to {self.tracefile}")
        if os.path.exists(self.tracefile):
            self.df.to_csv(self.tracefile, mode="a", header=False, index=False)
        else:
            self.df.to_csv(self.tracefile, mode="w", header=True, index=False)

    async def run_phase(self) -> None:
        """
        Run phase for the coverage component.

        """
        while True:
            # Wait for an item to be available on the port
            item = await self.item_port.blocking_get()

            # If no columns are defined, use all public attributes of the item
            if self.df is None:
                self.columns = []
                for k,v in vars(item).items():
                    if not k.startswith('_') and not callable(v):
                        if k not in self.columns:
                            self.columns.append(k)

                self.df = self.create_dataframe()

            # Create a row dictionary for the DataFrame
            # Populate the row with item attributes
            row = {}
            for col in self.df.columns:
                if hasattr(item, col):
                    row[col] = getattr(item, col)
                else:
                    row[col] = None

            if row['_timestamp_'] is None:
                row['_timestamp_'] = cocotb.utils.get_sim_time("ns")

            if row['_path_'] is None:
                row['_path_'] = self.get_full_name()

            # Append the row to the DataFrame
            self.df.loc[len(self.df)] = row

            # Flush and create new dataframe
            if len(self.df) >= self.flush_interval:
                self.flush()
                self.df = pd.DataFrame(columns=self.df.columns)

    async def report_phase(self) -> None:
        """
        Report phase for the bandwidth component.

        Generate plot of bytes on bus over time windows

        """
        self.flush()

__all__ = ["Trace"]



# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Port

from typing import Any

from .component import Component
from .list import List


class Port(Component):
    def __init__(self, name: str, parent: Component) -> None:
        """
        Initializes the port with a name and parent component.

        :param name: The name of the port.
        :type name: str
        :param parent: The parent component to which this port belongs.
        :type parent: Component
        """
        super().__init__(name, parent)
        self.channels = []

    def connect(self, dst_list: List) -> None:
        """
        Connects the port to a destination list.

        :param dst_list: The destination list to connect to.
        :type dst_list: List
        """
        self.channels.append(dst_list)

    def write(self, data: Any) -> None:
        """
        Writes data to the connected channels.

        :param data: The data to be written.
        """
        for c in self.channels:
            c.append(data)

    def delete(self) -> None:
        """
        Deletes the connected channels.
        """
        for c in self.channels:
            c.delete()


__all__ = ["Port"]



from .agent import Agent
from .bool import Bool
from .component import Component
from .coverage import Coverage
from .coverbin import Coverbin
from .covercross import Covercross
from .covergroup import Covergroup
from .coverpoint import Coverpoint
from .driver import Driver
from .enum import Enum
from .env import Env
from .factory import Factory
from .fifo import Fifo
from .float import Double, Float, Fp16, Fp32, Fp64, Half
from .int import Byte, Int, Int8, Int16, Int32, Int64
from .list import List, Queue
from .log import Log
from .logic import Logic
from .memory import Memory
from .model import Model
from .monitor import Monitor
from .object import Object
from .phase import Phase
from .phase_manager import PhaseManager
from .port import Port
from .scoreboard import Scoreboard
from .scoreboard_indexed import IndexedScoreboard
from .sequence import Sequence
from .sequence_item import SequenceItem
from .sequencer import Sequencer
from .struct import Struct
from .trace import Trace
from .transaction import Transaction
from .uint import Uint, Uint8, Uint16, Uint32, Uint64
from .visualization import Visualization

# Enable logging
Log._override_cocotb_logging()

# Add default phases
PhaseManager.add_phase("RUN", None, top_down=False)
PhaseManager.add_phase("REPORT", PhaseManager.get_phase("RUN"), top_down=False)

# Add version
__version__: str = "0.3.0"

__all__ = [
    "__version__",
    "Log",
    "List",
    "Queue",
    "Fifo",
    "Port",
    "Phase",
    "Phase_Manager",
    "Object",
    "Transaction",
    "Component",
    "Driver",
    "Sequencer",
    "Monitor",
    "Model",
    "Scoreboard",
    "IndexedScoreboard",
    "Agent",
    "Env",
    "SequenceItem",
    "Sequence",
    "Logic",
    "Bool",
    "Int",
    "Int8",
    "Byte",
    "Int16",
    "Int32",
    "Int64",
    "Uint",
    "Uint8",
    "Uint16",
    "Uint32",
    "Uint64",
    "Fp16",
    "Half",
    "Fp32",
    "Float",
    "Fp64",
    "Double",
    "Enum",
    "Coverbin",
    "Coverpoint",
    "Covercross",
    "Covergroup",
    "Coverage",
    "Visualization",
    "Factory",
    "Trace",
    "Memory",
    "Struct",
]



# Apheleia Verification Library Hierachical component
# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Coverpoint

from __future__ import annotations

from itertools import product
from typing import TYPE_CHECKING

from .coverbin import Coverbin
from .coverpoint import Coverpoint

if TYPE_CHECKING:
    from .covergroup import Covergroup

class Covercross(Coverpoint):
    def __init__(self, name: str, parent: Covergroup) -> None:
        """
        Initialize an instance of the class.

        :param name: The name of the instance
        :type name: str
        :param parent: The parent instance
        :type parent: object
        """
        super().__init__(name, parent, var=None)
        self._points_ = []

    def add_points(self, *args: list[Coverpoint]) -> None:
        """
        Add coverpoints to the cross.

        :param args: Coverpoints to be crossed
        :type args: list[Coverpoint]
        :raises ValueError: If the covercross already exists
        """
        self._points_ = args

        p = [arg._bins_ for arg in args]

        for c in product(*p):
            name = ", ".join([str(x) for x in c])
            self._bins_[name] = Coverbin(name, self, c)

    def sample(self) -> None:
        """
        Sample values from each bin in the `_bins_` attribute.

        This method iterates over all bins stored in the `_bins_` attribute and
        calls the `sample` method on each bin, passing the result of `self.var()`
        as an argument.
        """
        bins = []
        for p in self._points_:
            b = p.get_hit()
            if b is not None:
                bins.append(b.get_name())
            else:
                return

        v = tuple(bins)
        for b in self._bins_.values():
            b.sample(v)


__all__ = ["Covercross"]



# Copyright 2024 Apheleia
#
# Description:
# Apheleia Verification Library Indexed Scoreboard

from typing import Any

from .component import Component
from .scoreboard import Scoreboard


class IndexedScoreboard(Scoreboard):
    def __init__(self, name: str, parent: Component) -> None:
        """
        Initializes the indexed scoreboard component.

        :param name: Name of the indexed scoreboard.
        :type name: str
        :param parent: Parent component.
        :type parent: Component
        """
        super().__init__(name, parent)
        self.scoreboards = {}

    def set_indices(self, indices: list[Any]) -> None:
        """
        Sets the indices for the scoreboards.

        :param indices: List of indices.
        :type indices: list[Any]
        """
        for idx in indices:
            self.scoreboards[idx] = Scoreboard(f"{self.name}_{idx}", self)
            self.scoreboards[idx].set_verbose(self.verbose)
            self.scoreboards[idx].set_min_compare_count(self.min_compare_count)

    def set_verbose(self, verbose: bool) -> None:
        """
        Sets the verbosity of the scoreboards.

        :param verbose: Verbosity flag.
        :type verbose: bool
        """
        super().set_verbose(verbose)
        for v in self.scoreboards.values():
            v.set_verbose(verbose)

    def set_min_compare_count(self, count: int) -> None:
        """
        Sets the minimum number of comparisons required for the scoreboards.

        :param count: Minimum number of comparisons.
        :type count: int
        """
        super().set_min_compare_count(count)
        for v in self.scoreboards.values():
            v.set_min_compare_count(count)

    def get_index(self, item: Any) -> int:
        """
        Gets the index for the given item.

        :param item: The item to get the index for.
        :type item: Any
        :returns: The index of the item.
        :rtype: int
        """
        raise NotImplementedError

    async def run_phase(self) -> None:
        """
        Runs the scoreboard phase.
        """
        while True:
            self.before_item = await self.before_port.blocking_pop()
            self.scoreboards[self.get_index(self.before_item)].before_port.append(self.before_item)

            self.after_item = await self.after_port.blocking_pop()
            self.scoreboards[self.get_index(self.after_item)].after_port.append(self.after_item)

            self.compare_count += 1

            self.after_item = None
            self.before_item = None


__all__ = ["IndexedScoreboard"]
